"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var jsx_runtime_1 = require("react/jsx-runtime");
var react_1 = require("react");
var normalizeValue = function (min, value, max) { return Math.max(Math.min(value, max), min); };
var getStyleFabric = function (axis) {
    if (axis === void 0) { axis = 'y'; }
    return function (size) {
        var _a;
        return (_a = {},
            _a[axis === 'y' ? 'minHeight' : 'minWidth'] = size,
            _a[axis === 'y' ? 'height' : 'width'] = size,
            _a[axis === 'y' ? 'maxHeight' : 'maxWidth'] = size,
            _a.overflowAnchor = 'none',
            _a.pointerEvents = 'none',
            _a.userSelect = 'none',
            _a.padding = 0,
            _a.margin = 0,
            _a.border = 'none',
            _a);
    };
};
var MIN_INDEX = 0;
var IS_OVERFLOW_ANCHOR_SUPPORTED = (function () {
    try {
        return window.CSS.supports('overflow-anchor: auto');
    }
    catch (error) {
        return false;
    }
})();
var ViewportList = react_1.forwardRef(function (_a, ref) {
    var _b = _a.viewportRef, viewportRef = _b === void 0 ? null : _b, _c = _a.items, items = _c === void 0 ? [] : _c, itemMinSize = _a.itemMinSize, _d = _a.margin, margin = _d === void 0 ? 0 : _d, _e = _a.overscan, overscan = _e === void 0 ? 1 : _e, _f = _a.axis, axis = _f === void 0 ? 'y' : _f, _g = _a.initialIndex, initialIndex = _g === void 0 ? 0 : _g, _h = _a.initialAlignToTop, initialAlignToTop = _h === void 0 ? true : _h, children = _a.children;
    var _j = react_1.useMemo(function () { return ({
        propName: {
            top: axis === 'y' ? 'top' : 'left',
            bottom: axis === 'y' ? 'bottom' : 'right',
            clientHeight: axis === 'y' ? 'clientHeight' : 'clientWidth',
            scrollTop: axis === 'y' ? 'scrollTop' : 'scrollLeft',
            overflowY: axis === 'y' ? 'overflowY' : 'overflowX'
        },
        getStyle: getStyleFabric(axis)
    }); }, [axis]), propName = _j.propName, getStyle = _j.getStyle;
    var maxIndex = items.length - 1;
    var itemMinSizeWithMargin = itemMinSize + margin;
    var overscanSize = overscan * itemMinSizeWithMargin;
    var _k = react_1.useState(function () {
        var normalizedInitialIndex = normalizeValue(MIN_INDEX, initialIndex, maxIndex);
        return [normalizedInitialIndex, normalizedInitialIndex];
    }), _l = _k[0], startIndex = _l[0], endIndex = _l[1], setIndexes = _k[1];
    var topRef = react_1.useRef(null);
    var bottomRef = react_1.useRef(null);
    var variables = react_1.useRef({
        cache: [],
        step: function () { return null; },
        scrollToIndex: startIndex ? { index: startIndex, alignToTop: initialAlignToTop } : null,
        scrollCompensationEndIndex: null
    });
    var normalizedStartIndex = normalizeValue(MIN_INDEX, startIndex, maxIndex);
    var normalizedEndIndex = normalizeValue(normalizedStartIndex, endIndex, maxIndex);
    var topStyle = react_1.useMemo(function () {
        return getStyle(variables.current.cache
            .slice(MIN_INDEX, normalizedStartIndex)
            .reduce(function (sum, next) { return sum + next; }, normalizedStartIndex * itemMinSizeWithMargin));
    }, [normalizedStartIndex, itemMinSizeWithMargin, getStyle]);
    var bottomStyle = react_1.useMemo(function () {
        return getStyle(variables.current.cache
            .slice(normalizedEndIndex + 1, maxIndex)
            .reduce(function (sum, next) { return sum + next; }, itemMinSizeWithMargin * (maxIndex - normalizedEndIndex)));
    }, [normalizedEndIndex, maxIndex, itemMinSizeWithMargin, getStyle]);
    var slicedItems = react_1.useMemo(function () {
        return items
            .slice(normalizedStartIndex, normalizedEndIndex + 1)
            .map(function (item, index) { return children(item, normalizedStartIndex + index); });
    }, [items, normalizedStartIndex, normalizedEndIndex, children]);
    variables.current.step = function () {
        var _a, _b, _c, _d, _e, _f, _g;
        var viewportRect = (_a = viewportRef === null || viewportRef === void 0 ? void 0 : viewportRef.current) === null || _a === void 0 ? void 0 : _a.getBoundingClientRect();
        var topLimit = (viewportRect
            ? normalizeValue(0, viewportRect[propName.top], document.documentElement[propName.clientHeight])
            : 0) - overscanSize;
        var bottomLimit = (viewportRect
            ? normalizeValue(0, viewportRect[propName.bottom], document.documentElement[propName.clientHeight])
            : document.documentElement[propName.clientHeight]) + overscanSize;
        var topElementRect = (_c = (_b = topRef.current) === null || _b === void 0 ? void 0 : _b.nextSibling) === null || _c === void 0 ? void 0 : _c.getBoundingClientRect();
        var bottomElementRect = (_e = (_d = bottomRef.current) === null || _d === void 0 ? void 0 : _d.previousSibling) === null || _e === void 0 ? void 0 : _e.getBoundingClientRect();
        var maxItemsCountInViewPort = Math.ceil((bottomLimit - topLimit) / itemMinSizeWithMargin);
        var nextStartIndex = startIndex;
        var nextEndIndex = endIndex;
        var diff;
        var index;
        var element;
        if (!topElementRect || !bottomElementRect) {
            return;
        }
        if (variables.current.scrollToIndex) {
            var targetIndex = normalizeValue(MIN_INDEX, variables.current.scrollToIndex.index, maxIndex);
            if (targetIndex >= startIndex && targetIndex <= endIndex) {
                index = startIndex;
                element = (_f = topRef.current) === null || _f === void 0 ? void 0 : _f.nextSibling;
                while (element && element !== bottomRef.current) {
                    if (index === targetIndex) {
                        element.scrollIntoView(variables.current.scrollToIndex.alignToTop);
                        variables.current.scrollToIndex = null;
                        break;
                    }
                    element = element.nextSibling;
                    ++index;
                }
                return;
            }
            nextStartIndex = targetIndex - maxItemsCountInViewPort;
            nextEndIndex = targetIndex + maxItemsCountInViewPort;
        }
        else if (topElementRect[propName.top] >= bottomLimit) {
            diff = topElementRect[propName.top] - bottomLimit;
            nextEndIndex = startIndex;
            while (diff >= 0 && nextEndIndex > MIN_INDEX) {
                diff -= (variables.current.cache[--nextEndIndex] || 0) + itemMinSizeWithMargin;
            }
            variables.current.scrollCompensationEndIndex = startIndex;
            nextStartIndex = nextEndIndex - maxItemsCountInViewPort;
        }
        else if (bottomElementRect[propName.bottom] + margin <= topLimit) {
            diff = topLimit - bottomElementRect[propName.bottom] + margin;
            nextStartIndex = endIndex;
            while (diff >= 0 && nextStartIndex < maxIndex) {
                diff -= (variables.current.cache[++nextStartIndex] || 0) + itemMinSizeWithMargin;
            }
            nextEndIndex = nextStartIndex + maxItemsCountInViewPort;
        }
        else {
            if (topElementRect[propName.bottom] + margin < topLimit) {
                ++nextStartIndex;
            }
            else if (topElementRect[propName.top] >= topLimit) {
                diff = topElementRect[propName.top] - topLimit;
                while (diff >= 0 && nextStartIndex > MIN_INDEX) {
                    diff -= (variables.current.cache[--nextStartIndex] || 0) + itemMinSizeWithMargin;
                }
                variables.current.scrollCompensationEndIndex = startIndex;
            }
            if (bottomElementRect[propName.bottom] + margin <= bottomLimit) {
                diff = bottomLimit - bottomElementRect[propName.bottom] - margin;
                while (diff >= 0 && nextEndIndex < maxIndex) {
                    diff -= (variables.current.cache[++nextEndIndex] || 0) + itemMinSizeWithMargin;
                }
            }
            else if (bottomElementRect[propName.top] > bottomLimit) {
                --nextEndIndex;
            }
        }
        nextStartIndex = normalizeValue(MIN_INDEX, nextStartIndex, maxIndex);
        nextEndIndex = normalizeValue(nextStartIndex, nextEndIndex, maxIndex);
        if (nextStartIndex !== startIndex || nextEndIndex !== endIndex) {
            index = startIndex;
            element = (_g = topRef.current) === null || _g === void 0 ? void 0 : _g.nextSibling;
            while (!!element && index <= endIndex && element !== bottomRef.current) {
                variables.current.cache[index++] = element[propName.clientHeight] - itemMinSize;
                element = element.nextSibling;
            }
            setIndexes([nextStartIndex, nextEndIndex]);
        }
    };
    react_1.useImperativeHandle(ref, function () { return ({
        scrollToIndex: function (index, alignToTop) {
            if (index === void 0) { index = -1; }
            if (alignToTop === void 0) { alignToTop = true; }
            variables.current.scrollToIndex = { index: index, alignToTop: alignToTop };
        }
    }); }, []);
    react_1.useLayoutEffect(function () {
        var _a;
        if (variables.current.scrollCompensationEndIndex === null) {
            return;
        }
        if (!viewportRef || IS_OVERFLOW_ANCHOR_SUPPORTED) {
            variables.current.scrollCompensationEndIndex = null;
            return;
        }
        var index = startIndex;
        var element = (_a = topRef.current) === null || _a === void 0 ? void 0 : _a.nextSibling;
        var sizeDiff = 0;
        while (index < variables.current.scrollCompensationEndIndex && element !== bottomRef.current) {
            sizeDiff += element[propName.clientHeight] - (variables.current.cache[index] || 0) - itemMinSize;
            element = element.nextSibling;
            ++index;
        }
        if (viewportRef && sizeDiff) {
            viewportRef.current.style[propName.overflowY] = 'hidden';
            viewportRef.current[propName.scrollTop] += sizeDiff;
            viewportRef.current.style[propName.overflowY] = '';
        }
        variables.current.scrollCompensationEndIndex = null;
    }, [viewportRef, startIndex, itemMinSize, propName]);
    react_1.useEffect(function () {
        var frameId;
        var frame = function () {
            frameId = requestAnimationFrame(frame);
            variables.current.step();
        };
        frame();
        return function () {
            cancelAnimationFrame(frameId);
        };
    }, []);
    return (jsx_runtime_1.jsxs(react_1.Fragment, { children: [jsx_runtime_1.jsx("div", { ref: topRef, style: topStyle }, "ViewPortListTop"), slicedItems, jsx_runtime_1.jsx("div", { ref: bottomRef, style: bottomStyle }, "ViewPortListBottom")] }, "ViewPortList"));
});
exports.default = ViewportList;
//# sourceMappingURL=index.js.map